@@ -36,10 +36,10 @@ from PySide.QtCore import QT_TRANSLATE_NOOP
 
 import FreeCAD as App
 import FreeCADGui as Gui
-
-from draftutils.messages import _msg
-from draftutils.translate import translate
 from draftobjects.layer import Layer
+from draftutils import params
+from draftutils import utils
+from draftutils.translate import translate
 
 
 class ViewProviderLayer:
@@ -70,16 +70,16 @@ class ViewProviderLayer:
                              _tip)
             vobj.OverrideLineColorChildren = True
 
-        if "OverrideShapeColorChildren" not in properties:
+        if "OverrideShapeAppearanceChildren" not in properties:
             _tip = QT_TRANSLATE_NOOP("App::Property",
                                      "If it is true, the objects contained "
                                      "within this layer will adopt "
-                                     "the shape color of the layer")
+                                     "the shape appearance of the layer")
             vobj.addProperty("App::PropertyBool",
-                             "OverrideShapeColorChildren",
+                             "OverrideShapeAppearanceChildren",
                              "Layer",
                              _tip)
-            vobj.OverrideShapeColorChildren = True
+            vobj.OverrideShapeAppearanceChildren = True
 
         if "UsePrintColor" not in properties:
             _tip = QT_TRANSLATE_NOOP("App::Property",
@@ -91,11 +91,8 @@ class ViewProviderLayer:
                              "Print",
                              _tip)
 
-
     def set_visual_properties(self, vobj, properties):
         """Set visual properties only if they don't already exist."""
-        view_group = App.ParamGet("User parameter:BaseApp/Preferences/View")
-
         if "LineColor" not in properties:
             _tip = QT_TRANSLATE_NOOP("App::Property",
                                      "The line color of the objects "
@@ -104,11 +101,7 @@ class ViewProviderLayer:
                              "LineColor",
                              "Layer",
                              _tip)
-
-            c = view_group.GetUnsigned("DefaultShapeLineColor", 255)
-            vobj.LineColor = (((c >> 24) & 0xFF) / 255,
-                              ((c >> 16) & 0xFF) / 255,
-                              ((c >> 8) & 0xFF) / 255)
+            vobj.LineColor = params.get_param_view("DefaultShapeLineColor") & 0xFFFFFF00
 
         if "ShapeColor" not in properties:
             _tip = QT_TRANSLATE_NOOP("App::Property",
@@ -117,12 +110,21 @@ class ViewProviderLayer:
             vobj.addProperty("App::PropertyColor",
                              "ShapeColor",
                              "Layer",
-                             _tip)
+                             _tip,
+                             4)  # Hidden
+            vobj.ShapeColor = params.get_param_view("DefaultShapeColor") & 0xFFFFFF00
 
-            c = view_group.GetUnsigned("DefaultShapeColor", 4294967295)
-            vobj.ShapeColor = (((c >> 24) & 0xFF) / 255,
-                               ((c >> 16) & 0xFF) / 255,
-                               ((c >> 8) & 0xFF) / 255)
+        if "ShapeAppearance" not in properties:
+            _tip = QT_TRANSLATE_NOOP("App::Property",
+                                     "The shape appearance of the objects "
+                                     "contained within this layer")
+            vobj.addProperty("App::PropertyMaterialList",
+                             "ShapeAppearance",
+                             "Layer",
+                             _tip)
+            material = App.Material()
+            material.DiffuseColor = params.get_param_view("DefaultShapeColor") & 0xFFFFFF00
+            vobj.ShapeAppearance = (material, )
 
         if "LineWidth" not in properties:
             _tip = QT_TRANSLATE_NOOP("App::Property",
@@ -132,9 +134,7 @@ class ViewProviderLayer:
                              "LineWidth",
                              "Layer",
                              _tip)
-
-            width = view_group.GetInt("DefaultShapeLineWidth", 2)
-            vobj.LineWidth = width
+            vobj.LineWidth = params.get_param_view("DefaultShapeLineWidth")
 
         if "DrawStyle" not in properties:
             _tip = QT_TRANSLATE_NOOP("App::Property",
@@ -144,8 +144,8 @@ class ViewProviderLayer:
                              "DrawStyle",
                              "Layer",
                              _tip)
-            vobj.DrawStyle = ["Solid", "Dashed", "Dotted", "Dashdot"]
-            vobj.DrawStyle = "Solid"
+            vobj.DrawStyle = utils.DRAW_STYLES
+            vobj.DrawStyle = params.get_param("DefaultDrawStyle")
 
         if "Transparency" not in properties:
             _tip = QT_TRANSLATE_NOOP("App::Property",
@@ -155,7 +155,7 @@ class ViewProviderLayer:
                              "Transparency",
                              "Layer",
                              _tip)
-            vobj.Transparency = 0
+            vobj.Transparency = params.get_param_view("DefaultShapeTransparency")
 
         if "LinePrintColor" not in properties:
             _tip = QT_TRANSLATE_NOOP("App::Property",
@@ -166,6 +166,7 @@ class ViewProviderLayer:
                              "LinePrintColor",
                              "Print",
                              _tip)
+            vobj.LinePrintColor = params.get_param("DefaultPrintColor")
 
     def getIcon(self):
         """Return the path to the icon used by the viewprovider.
@@ -189,6 +190,8 @@ class ViewProviderLayer:
         These are the elements of the `Group` property of the Proxy object.
         """
         if hasattr(self, "Object") and hasattr(self.Object, "Group"):
+            if getattr(self.Object.ViewObject, "HideChildren", False):
+                return []
             return self.Object.Group
 
     def getDisplayModes(self, vobj):
@@ -203,19 +206,19 @@ class ViewProviderLayer:
         """Return the saved display mode."""
         return mode
 
-    def __getstate__(self):
+    def dumps(self):
         """Return a tuple of objects to save or None."""
         return None
 
-    def __setstate__(self, state):
+    def loads(self, state):
         """Set the internal properties from the restored state."""
         return None
 
     def updateData(self, obj, prop):
         """Execute when a property from the Proxy class is changed."""
         if prop == "Group":
-            for _prop in ("LineColor", "ShapeColor", "LineWidth",
-                          "DrawStyle", "Transparency", "Visibility"):
+            for _prop in ("LineColor", "ShapeAppearance", "LineWidth",
+                          "DrawStyle", "Visibility"):
                 self.onChanged(obj.ViewObject, _prop)
 
     def change_view_properties(self, vobj, prop):
@@ -225,16 +228,15 @@ class ViewProviderLayer:
         # Return if the property does not exist
         if not hasattr(vobj, prop):
             return
+        # If the override properties are not set return without change
+        if prop == "LineColor" and not vobj.OverrideLineColorChildren:
+            return
+        elif prop == "ShapeAppearance" and not vobj.OverrideShapeAppearanceChildren:
+            return
 
         for target_obj in obj.Group:
             target_vobj = target_obj.ViewObject
 
-            # If the override properties are not set return without change
-            if prop == "LineColor" and not vobj.OverrideLineColorChildren:
-                return
-            elif prop == "ShapeColor" and not vobj.OverrideShapeColorChildren:
-                return
-
             # This checks that the property exists in the target object,
             # and then sets the target property accordingly
             if hasattr(target_vobj, prop):
@@ -253,18 +255,49 @@ class ViewProviderLayer:
 
     def onChanged(self, vobj, prop):
         """Execute when a view property is changed."""
-        if prop in ("LineColor", "ShapeColor", "LineWidth",
-                    "DrawStyle", "Transparency", "Visibility"):
+
+        if not hasattr(vobj, prop):
+            return
+
+        if prop == "ShapeColor":
+            if hasattr(vobj, "ShapeAppearance"):
+                material = vobj.ShapeAppearance[0]
+                if material.DiffuseColor != vobj.ShapeColor:
+                    material.DiffuseColor = vobj.ShapeColor
+                    vobj.ShapeAppearance = (material, )
+            # The changed ShapeAppearance will do the rest:
+            return
+
+        if prop == "Transparency":
+            if hasattr(vobj, "ShapeAppearance"):
+                material = vobj.ShapeAppearance[0]
+                if material.Transparency != vobj.Transparency / 100:
+                    material.Transparency = vobj.Transparency / 100
+                    vobj.ShapeAppearance = (material, )
+            # The changed ShapeAppearance will do the rest:
+            return
+
+        if (prop == "ShapeAppearance"
+                and hasattr(vobj, "ShapeColor")
+                and hasattr(vobj, "Transparency")):
+            material = vobj.ShapeAppearance[0]
+            if material.DiffuseColor != vobj.ShapeColor:
+                vobj.ShapeColor = material.DiffuseColor
+            if material.Transparency != vobj.Transparency / 100:
+                vobj.Transparency = int(material.Transparency * 100)
+
+        if (prop in ("LineColor", "ShapeAppearance", "LineWidth",
+                     "DrawStyle", "Visibility")
+                and hasattr(vobj, "OverrideLineColorChildren")
+                and hasattr(vobj, "OverrideShapeAppearanceChildren")):
             self.change_view_properties(vobj, prop)
 
-        if (prop in ("LineColor", "ShapeColor")
+        # Paint the layer icon in the tree view:
+        if (prop in ("LineColor", "ShapeAppearance")
                 and hasattr(vobj, "LineColor")
-                and hasattr(vobj, "ShapeColor")):
-            # This doesn't do anything to the objects inside the layer,
-            # it just uses the defined Line and Shape colors
-            # to paint the layer icon accordingly in the tree view
+                and hasattr(vobj, "ShapeAppearance")):
             l_color = vobj.LineColor
-            s_color = vobj.ShapeColor
+            s_color = vobj.ShapeAppearance[0].DiffuseColor
 
             l_color = QtGui.QColor(int(l_color[0] * 255),
                                    int(l_color[1] * 255),
@@ -296,8 +329,40 @@ class ViewProviderLayer:
             self.icondata = byte_array.data().decode("latin1")
             vobj.signalChangeIcon()
 
+    def _get_layer(self, obj):
+        """Get the layer the object belongs to.
+        """
+        from draftmake.make_layer import get_layer_container
+        # First look in the LayerContainer:
+        for layer in get_layer_container().Group:
+            if utils.get_type(layer) == "Layer" and obj in layer.Group:
+                return layer
+        # If not found, look through all App::FeaturePython objects (not just layers):
+        for find in obj.Document.findObjects(Type="App::FeaturePython"):
+            if utils.get_type(find) == "Layer" and obj in find.Group:
+                return find
+        return None
+
     def canDragObject(self, obj):
-        """Return True to allow dragging one object from the Layer."""
+        """Return True to allow dragging one object from the Layer.
+
+        Also store parent group data for update_groups_after_drag_drop and
+        trigger that function.
+        """
+        if not hasattr(self, "old_parent_data"):
+            self.old_parent_data = {}
+        old_data = []
+        for parent in obj.InList:
+            if hasattr(parent, "Group"):
+                old_data.append([parent, parent.Group])
+        # Layers are not in the Inlist because a layer's Group is App::PropertyLinkListHidden:
+        layer = self._get_layer(obj)
+        if layer is not None:
+            old_data.append([layer, layer.Group])
+        if old_data:
+            self.old_parent_data.setdefault(obj, old_data)
+            QtCore.QTimer.singleShot(0, self.update_groups_after_drag_drop)
+
         return True
 
     def canDragObjects(self):
@@ -306,10 +371,11 @@ class ViewProviderLayer:
 
     def dragObject(self, vobj, otherobj):
         """Remove the object that was dragged from the layer."""
-        if hasattr(vobj.Object, "Group") and otherobj in vobj.Object.Group:
-            group = vobj.Object.Group
+        layer = vobj.Object
+        if otherobj in layer.Group:
+            group = layer.Group
             group.remove(otherobj)
-            vobj.Object.Group = group
+            layer.Group = group
             App.ActiveDocument.recompute()
 
     def canDropObject(self, obj):
@@ -317,10 +383,13 @@ class ViewProviderLayer:
 
         If the object being dropped is itself a `'Layer'`, return `False`
         to prevent dropping a layer inside a layer, at least for now.
+
+        Also store parent group data for update_groups_after_drag_drop and
+        trigger that function.
         """
-        if hasattr(obj, "Proxy") and isinstance(obj.Proxy, Layer):
+        if utils.get_type(obj) == "Layer":
             return False
-        return True
+        return self.canDragObject(obj)
 
     def canDropObjects(self):
         """Return true to allow dropping many objects."""
@@ -333,55 +402,120 @@ class ViewProviderLayer:
         return immediately to prevent dropping a layer inside a layer,
         at least for now.
         """
-        if hasattr(otherobj, "Proxy") and isinstance(otherobj.Proxy, Layer):
+        if utils.get_type(otherobj) == "Layer":
             return
 
-        obj = vobj.Object
+        # We assume a single old layer...
 
-        if hasattr(obj, "Group") and otherobj not in obj.Group:
-            group = obj.Group
+        old_layer = self._get_layer(otherobj)
+        if old_layer is not None:
+            group = old_layer.Group
+            group.remove(otherobj)
+            old_layer.Group = group
+
+        new_layer = vobj.Object
+        if otherobj not in new_layer.Group:
+            group = new_layer.Group
             group.append(otherobj)
-            obj.Group = group
+            new_layer.Group = group
 
-            # Remove from all other layers (not automatic)
-            for parent in otherobj.InList:
-                if (hasattr(parent, "Proxy")
-                        and isinstance(parent.Proxy, Layer)
-                        and otherobj in parent.Group
-                        and parent != obj):
-                    p_group = parent.Group
-                    p_group.remove(otherobj)
-                    parent.Group = p_group
+        App.ActiveDocument.recompute()
 
+    def update_groups_after_drag_drop(self):
+        """Workaround function to improve the drag and drop behavior of Layer
+        objects.
+
+        The function processes the parent group data stored in the
+        old_parent_data dictionary by canDragObject and canDropObject.
+        """
+
+        # The function can be called multiple times, old_parent_data will be
+        # empty after the first call.
+        if (not hasattr(self, "old_parent_data")) or (not self.old_parent_data):
+            return
+
+        # List to collect parents whose Group must be updated.
+        # This has to happen later in a separate loop as we need the unmodified
+        # InList properties of the children in the main loop.
+        parents_to_update = []
+
+        # Main loop:
+        for child, old_data in self.old_parent_data.items():
+
+            # We assume a single old and a single new layer...
+
+            old_layer = None
+            for old_parent, old_parent_group in old_data:
+                if utils.get_type(old_parent) == "Layer":
+                    old_layer = old_parent
+                    break
+
+            new_layer = self._get_layer(child)
+            if new_layer == old_layer:
+                continue
+
+            elif new_layer is None:
+                # An object was dragged out of a layer.
+                # We need to check if it was put in a new group. If that is
+                # the case the content of old_layer should be restored.
+                # If the object was not put in a new group it was dropped on
+                # the document node, in that case we do nothing.
+                old_parents = [sub[0] for sub in old_data]
+                for new_parent in child.InList:
+                    if (hasattr(new_parent, "Group")
+                            and new_parent not in old_parents): # New group check.
+                        for old_parent, old_parent_group in old_data:
+                            if old_parent == old_layer:
+                                parents_to_update.append([old_parent, old_parent_group])
+                                break
+                        break
+
+            else:
+                # A new layer was assigned.
+                # The content of all `non-layer` groups should be restored.
+                for old_parent, old_parent_group in old_data:
+                    if utils.get_type(old_parent) != "Layer":
+                        parents_to_update.append([old_parent, old_parent_group])
+
+        # Update parents:
+        if parents_to_update:
+            for old_parent, old_parent_group in parents_to_update:
+                old_parent.Group = old_parent_group
             App.ActiveDocument.recompute()
 
+        self.old_parent_data = {}
+
+    def replaceObject(self, old_obj, new_obj):
+        """Return immediately to prevent replacement of children."""
+        return
+
     def setupContextMenu(self, vobj, menu):
         """Set up actions to perform in the context menu."""
-        action1 = QtGui.QAction(QtGui.QIcon(":/icons/button_right.svg"),
-                                translate("draft", "Activate this layer"),
-                                menu)
-        action1.triggered.connect(self.activate)
-        menu.addAction(action1)
+        action_activate = QtGui.QAction(QtGui.QIcon(":/icons/button_right.svg"),
+                                        translate("draft", "Activate this layer"),
+                                        menu)
+        action_activate.triggered.connect(self.activate)
+        menu.addAction(action_activate)
 
-        action2 = QtGui.QAction(QtGui.QIcon(":/icons/Draft_SelectGroup.svg"),
-                                translate("draft", "Select layer contents"),
-                                menu)
-        action2.triggered.connect(self.select_contents)
-        menu.addAction(action2)
+        action_select = QtGui.QAction(QtGui.QIcon(":/icons/Draft_SelectGroup.svg"),
+                                      translate("draft", "Select layer contents"),
+                                      menu)
+        action_select.triggered.connect(self.select_contents)
+        menu.addAction(action_select)
 
     def activate(self):
         """Activate the selected layer, it becomes the Autogroup."""
-        if hasattr(self, "Object"):
-            Gui.Selection.clearSelection()
-            Gui.Selection.addSelection(self.Object)
-            Gui.runCommand("Draft_AutoGroup")
+        Gui.Selection.clearSelection()
+        Gui.Selection.addSelection(self.Object)
+        if not "Draft_AutoGroup" in Gui.listCommands():
+            Gui.activateWorkbench("DraftWorkbench")
+        Gui.runCommand("Draft_AutoGroup")
 
     def select_contents(self):
         """Select the contents of the layer."""
-        if hasattr(self, "Object"):
-            Gui.Selection.clearSelection()
-            for layer_obj in self.Object.Group:
-                Gui.Selection.addSelection(layer_obj)
+        Gui.Selection.clearSelection()
+        for layer_obj in self.Object.Group:
+            Gui.Selection.addSelection(layer_obj)
 
 
 class ViewProviderLayerContainer:
@@ -401,29 +535,27 @@ class ViewProviderLayerContainer:
 
     def setupContextMenu(self, vobj, menu):
         """Set up actions to perform in the context menu."""
-        action1 = QtGui.QAction(QtGui.QIcon(":/icons/Draft_Layer.svg"),
-                                translate("draft", "Merge layer duplicates"),
-                                menu)
-        action1.triggered.connect(self.merge_by_name)
-        menu.addAction(action1)
-        action2 = QtGui.QAction(QtGui.QIcon(":/icons/Draft_NewLayer.svg"),
-                                translate("draft", "Add new layer"),
-                                menu)
-        action2.triggered.connect(self.add_layer)
-        menu.addAction(action2)
+        action_merge = QtGui.QAction(QtGui.QIcon(":/icons/Draft_Layer.svg"),
+                                     translate("draft", "Merge layer duplicates"),
+                                     menu)
+        action_merge.triggered.connect(self.merge_by_name)
+        menu.addAction(action_merge)
+
+        action_add = QtGui.QAction(QtGui.QIcon(":/icons/Draft_NewLayer.svg"),
+                                   translate("draft", "Add new layer"),
+                                   menu)
+        action_add.triggered.connect(self.add_layer)
+        menu.addAction(action_add)
 
     def merge_by_name(self):
         """Merge the layers that have the same base label."""
-        if not hasattr(self, "Object") or not hasattr(self.Object, "Group"):
-            return
-
         doc = App.ActiveDocument
         doc.openTransaction(translate("draft", "Merge layer duplicates"))
 
         layer_container = self.Object
         layers = []
         for obj in layer_container.Group:
-            if hasattr(obj, "Proxy") and isinstance(obj.Proxy, Layer):
+            if utils.get_type(obj) == "Layer":
                 layers.append(obj)
 
         to_delete = []
@@ -450,12 +582,9 @@ class ViewProviderLayerContainer:
                     base.Group = base_group
                 to_delete.append(layer)
             elif layer.Label != base_label:
-                _msg(translate("draft", "Relabeling layer:")
-                        + " '{}' -> '{}'".format(layer.Label, base_label))
                 layer.Label = base_label
 
         for layer in to_delete:
-            _msg(translate("draft", "Merging layer:") + " '{}'".format(layer.Label))
             doc.removeObject(layer.Name)
 
         doc.recompute()
@@ -468,19 +597,24 @@ class ViewProviderLayerContainer:
         doc = App.ActiveDocument
         doc.openTransaction(translate("draft", "Add new layer"))
 
-        Draft.make_layer()
+        Draft.make_layer(name=None, line_color=None, shape_color=None,
+                         line_width=None, draw_style=None, transparency=None)
 
         doc.recompute()
         doc.commitTransaction()
 
-    def __getstate__(self):
+    def dumps(self):
         """Return a tuple of objects to save or None."""
         return None
 
-    def __setstate__(self, state):
+    def loads(self, state):
         """Set the internal properties from the restored state."""
         return None
 
+    def replaceObject(self, old_obj, new_obj):
+        """Return immediately to prevent replacement of children."""
+        return
+
 
 # Alias for compatibility with v0.18 and earlier
 _ViewProviderVisGroup = ViewProviderLayer
